/* -*- c++ -*- */
/* 
 * Copyright 2015 ELAD .
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <libusb-1.0/libusb.h>

#define S_RATE 122880000

#include <gnuradio/io_signature.h>
#include "fdmduo_source_c_impl.h"

#define ZeroMemory(p,l)			memset(p, 0x00, l)
#define ZERO_MEMORY(p)			ZeroMemory(&p, sizeof(p))
#define SAFE_DELETE_ARRAY(p)	{ if (p) { delete [] p; p = NULL; } }
#define SAFE_DELETE(p)			{ if (p) { delete p; p = NULL; } }

namespace gr {
	namespace elad {

		typedef struct _cbdata_t {
			fdmduo_source_c_impl *obj;
			struct libusb_transfer *transfer;
		} cbdata_t, *cbdata_p;

		fdmduo_source_c::sptr fdmduo_source_c::make() {
			return gnuradio::get_initial_sptr (new fdmduo_source_c_impl());
		}

                fdmduo_source_c_impl::fdmduo_source_c_impl()
                        :gr::sync_block("fdmduo_source_c", gr::io_signature::make(0, 0, 0), gr::io_signature::make(1, 1, 2*sizeof(float))),
			m_nBufferItems(0),
			m_pUSBBuffer(NULL),
			m_nBufferStart(0),
			m_nBufferSize(512*24*8),
			m_bBuffering(false),
			m_nReadLength(DEFAULT_READLEN),
			m_nBufferMultiplier(DEFAULT_BUFFER_MUL),
			m_fBufferLevel(DEFAULT_BUFFER_LEVEL),
			m_nReadPacketCount(0),
			m_nBufferOverflowCount(0),
			m_nBufferUnderrunCount(0),
			m_bytes_per_sample(4)
			{	

			int n=0;

			m_nReadLength = 512*24;
			m_nBufferMultiplier = 8;
			m_bytes_per_sample = 4;

			m_bRunning = false;

			libusb_context *ctx;
			libusb_device_handle *dev_handle; 

			int res;
			unsigned char buffer[1024];
			int j;
			long freq;
			long tune;
			double tuningFreq;
			int sampleRateCorr;
			unsigned int tuningWordHex, twLS, twMS;

                        // Must be the same since rate is determined by the libusb reads!
                        m_recv_samples_per_packet = m_nReadLength;

                        m_nBufferSize = m_recv_samples_per_packet * m_nBufferMultiplier;
                        m_pUSBBuffer = new uint8_t[m_nBufferSize];
                        assert(m_pUSBBuffer);
                        memset(m_pUSBBuffer,0, m_nBufferSize);
                        calc_delay(19200);

			fprintf( stderr, "Operations init\n" );

			// Initialize libusb-1.0
			res = libusb_init(&ctx);
			if( res < 0 ){
				fprintf( stderr, "Init Error %d\n", res ); 
			}

			// set verbosity level to 3, as suggested in the documentation
			libusb_set_debug( ctx, 3 );

			dev_handle = libusb_open_device_with_vid_pid( ctx, 0x1721, 0x061a );
			if( dev_handle == NULL ){
				fprintf( stderr, "Cannot open FDM DUO Device\n" ); 
			} else {
				fprintf( stderr, "FDM DUO Device Opened\n" ); 
			}

			// detach kernel drivers
			if( libusb_kernel_driver_active( dev_handle, 0 ) == 1 ) {
				fprintf( stderr, "Kernel Driver Active\n" ); 
				if( libusb_detach_kernel_driver( dev_handle, 0 ) == 0 ) {
					fprintf( stderr, "Kernel Driver Detached\n" ); 
				}
			}

			// claim interface
			if( libusb_claim_interface( dev_handle, 0 ) < 0 ) {
				fprintf( stderr, "Cannot claim Interface\n" ); 
			}
			fprintf( stderr, "Interface claimed\n" ); 

			// read USB driver version
			memset( buffer, 0, sizeof( buffer ) );
			res = libusb_control_transfer( dev_handle, 0xc0, 0xFF, 0x0000, 0x0000, buffer, 2, 0 );
			if( res  ){
				fprintf( stderr, "USB Driver Version: %d.%d\n", buffer[0], buffer[1] ); 
			} else {
				fprintf( stderr, "USB Driver Version read failed\n" ); 
			}

			// read HW version
			memset( buffer, 0, sizeof( buffer ) );
			res = libusb_control_transfer( dev_handle, 0xc0, 0xA2, 0x404C, 0x0151, buffer, 2, 0 );
			if( res==2  ){
				fprintf( stderr, "HW Version: %d.%d\n", buffer[0], buffer[1] ); 
			} else {
				fprintf( stderr, "HW Version read failed\n" ); 
			}

			// read serial number
			memset( buffer, 0, sizeof( buffer ) );
			res = libusb_control_transfer( dev_handle, 0xc0, 0xA2, 0x4000, 0x0151, buffer, 32, 0 );
			if( res==32  ){
				fprintf( stderr, "Serial: %-s\n", buffer ); 
			} else {
				fprintf( stderr, "Serial read failed (%d)\n", res ); 
			}

			// stop FIFO
			memset( buffer, 0, sizeof( buffer ) );
			res = libusb_control_transfer( dev_handle, 0xc0, 0xE1, 0x0000, 0xE9<<8, buffer, 1, 0 );
			// res = control_transfer( localDeviceInfo.winUsbHandle, 0xc0, 0xE1, 0x0000, 0x00E9<<8, buffer, 1, 0 );
			if( res==1  ){
				fprintf( stderr, "Stop FIFO\n" ); 
			} else {
				fprintf( stderr, "Stop FIFO failed (%d)\n", res ); 
			}

			// imposta la FIFO di EP6 del CY-USB in modalità slave...
			memset( buffer, 0, sizeof( buffer ) );
			res = libusb_control_transfer( dev_handle, 0xc0, 0xE1, 0x0000, 0xE8<<8, buffer, 1, 0 );
			// res = control_transfer( localDeviceInfo.winUsbHandle, 0xc0, 0xE1, 0x0000, 0x00E8<<8, buffer, 1, 0 );
			if( res==1  ){
				fprintf( stderr, "Init FIFO\n" ); 
			} else {
				fprintf( stderr, "Init FIFO failed (%d)\n", res ); 
			}

			// start FIFO
			res = libusb_control_transfer( dev_handle, 0xc0, 0xE1, 0x0001, 0x0E9<<8, buffer, 1, 0 );
			if( res != 1 || buffer[0] != 0xE9 ) {
				fprintf( stderr, "Enable SYN failed (%d %02X)\n", res, buffer[0] ); 
			}
			fprintf( stderr, "SYN Enabled\n" );

			myDevInfo.deviceHandle = dev_handle;				
		}

		fdmduo_source_c_impl::~fdmduo_source_c_impl() {
			stop();
		}

		bool fdmduo_source_c_impl::start() {
			boost::recursive_mutex::scoped_lock lock(d_mutex);

			struct sched_param param;
			param.sched_priority=90;

			if (m_bRunning) {
				return true;
			}

			// Need to set this BEFORE starting thread (otherwise it will exit)
			m_bRunning = true;				

			m_bBuffering = true;
			m_pCaptureThread = boost::thread(_capture_thread, this);

			pthread_t thread_id=(pthread_t)m_pCaptureThread.native_handle();
			// if you need priority, decomment the following instruction
			// pthread_setschedparam( thread_id, SCHED_RR, &param );

			return true;
		}

		bool fdmduo_source_c_impl::stop() {

			boost::recursive_mutex::scoped_lock lock(d_mutex);

			if (m_bRunning == false) {
				return true;
			}

			// Must set before 'join' as this will signal capture thread to exit
			m_bRunning = false;			

			// In case general_work is waiting
			m_hPacketEvent.notify_one();	
			// Release lock AFTER notification, so waiting thread resumes in lockstep
			lock.unlock();					
			// Wait for capture thread to finish
			m_pCaptureThread.join();		

			return true;
		}

		void fdmduo_source_c_impl::calc_delay(long dSampleRate) {

			boost::recursive_mutex::scoped_lock lock(d_mutex);

			double dDelay = 1000000000ULL * WAIT_FUDGE / (double)((dSampleRate) / (double)m_nReadLength);
			uint64_t delay = (uint64_t)ceil(dDelay);

			m_wait_delay.sec = delay / 1000000000ULL;
			m_wait_delay.nsec = delay % 1000000000ULL;
		}

		void fdmduo_source_c_impl::_capture_thread(fdmduo_source_c_impl *p) {
			return p->capture_thread();
		}

		void fdmduo_source_c_impl::fdmduo_cb_in( struct libusb_transfer * transfer ) {

			boost::recursive_mutex::scoped_lock lock(d_mutex, boost::defer_lock);

			// printf( "Callback\n!" );

			cbdata_p cbd = (cbdata_p)transfer->user_data;
			struct libusb_transfer * transfer_in = cbd->transfer;

			libusb_submit_transfer( transfer_in );

			uint8_t* pBuffer = transfer->buffer;

			uint32_t nRemaining = std::min(m_nBufferSize - m_nBufferItems, (uint32_t)m_nReadLength);
			bool bSignal = true;
		
			lock.lock();

			if (nRemaining > 0) {
				uint8_t* p = m_pUSBBuffer + ((m_nBufferStart + m_nBufferItems) % m_nBufferSize);
				uint32_t nPart1 = (m_nBufferStart + m_nBufferItems) % m_nBufferSize;
				uint32_t nSize1 = std::min(nRemaining, m_nBufferSize - nPart1);
				memcpy(p, pBuffer, nSize1);

				uint32_t nResidual = nRemaining - nSize1;

				if (nResidual > 0) {
					memcpy(m_pUSBBuffer, pBuffer + nSize1, nResidual);
				}

				m_nBufferItems += nRemaining;

				if ((m_bBuffering) && (m_nBufferItems >= (uint32_t)(m_recv_samples_per_packet + (float)m_nBufferSize * m_fBufferLevel))) {
					std::cout << "Finished buffering: " << "m_nBufferItems: " << m_nBufferItems << 
							", m_nBufferSize: " << m_nBufferSize << 
							", m_nReadPacketCount: " << m_nReadPacketCount 
							<< std::endl << std::endl;

					m_bBuffering = false;
				}
				bSignal = !m_bBuffering;
			} else {
				std::cout << "rB" << std::endl;
				++m_nBufferOverflowCount;
			}
	
			lock.unlock();
	
			if (bSignal) {
				m_hPacketEvent.notify_one();
			}
		}

		void cb_in( struct libusb_transfer * transfer ) {

			cbdata_p cbd = (cbdata_p)transfer->user_data;
			fdmduo_source_c_impl *obj = cbd->obj;
			obj->fdmduo_cb_in( transfer );
		}

		void fdmduo_source_c_impl::capture_thread() {

			boost::recursive_mutex::scoped_lock lock(d_mutex, boost::defer_lock);

			if ((m_nReadLength == 0) || (m_pUSBBuffer == NULL) || (m_nBufferSize == 0) || (m_fBufferLevel < 0)) {
				lock.lock();
				// This will signal EOF
				m_bRunning = false;				
				m_hPacketEvent.notify_one();
				lock.unlock();

				std::cout << "Capture threading NOT starting due to state error: " 
						<< boost::this_thread::get_id() << std::endl;
				std::cout << 	"m_nReadLength "<< m_nReadLength <<
						" m_pUSBBuffer " << m_pUSBBuffer << 
						" m_nBufferSize " << m_nBufferSize << 
						" m_fBufferLevel " << m_fBufferLevel << std::endl;
				return;
			}
	  
			std::cout << "\nCapture threading starting: " << boost::this_thread::get_id() << std::endl;
	  
			uint8_t* pBuffer1 = new uint8_t[m_nReadLength];
			uint8_t* pBuffer2 = new uint8_t[m_nReadLength];
	  
			struct libusb_transfer * transfer_in1 = libusb_alloc_transfer(0);
			struct libusb_transfer * transfer_in2 = libusb_alloc_transfer(0);
			
			cbdata_t cbdata1;
			cbdata_t cbdata2;

			cbdata1.obj=this;
			cbdata1.transfer=transfer_in1;
			cbdata2.obj=this;
			cbdata2.transfer=transfer_in2;

			libusb_fill_bulk_transfer( transfer_in1, myDevInfo.deviceHandle, 0x86, pBuffer1, m_nReadLength, cb_in, &cbdata1, 5000);
			libusb_fill_bulk_transfer( transfer_in2, myDevInfo.deviceHandle, 0x86, pBuffer2, m_nReadLength, cb_in, &cbdata2, 5000);
			libusb_submit_transfer( transfer_in1 );
			libusb_submit_transfer( transfer_in2 );

			std::cout << "Capture threading started: " << std::endl;
			
			for( ;; ) {
				lock.lock();
				if (m_bRunning == false) {
					break;
				}
				lock.unlock();
				int r=libusb_handle_events_completed( NULL, NULL );
			}
	  
			std::cout << "Capture threading exiting: " << boost::this_thread::get_id() << std::endl;
capture_thread_exit:
			std::cout << "Clear pBuffer... done." << std::endl;			
			SAFE_DELETE_ARRAY(pBuffer1);
			SAFE_DELETE_ARRAY(pBuffer2);
		}



		int fdmduo_source_c_impl::work(int noutput_items_c, gr_vector_const_void_star &input_items, gr_vector_void_star &output_items) {
			float *out = (float *) output_items[0];
			int noutput_items=noutput_items_c*2;

			boost::recursive_mutex::scoped_lock lock(d_mutex);

			if (m_bRunning == false) {
				std::cout << "Work called while not running!" << std::endl;
				return -1;	// EOF
			}
	  
			if (noutput_items*m_bytes_per_sample > m_nBufferSize) {
				std::cout << "Work wants ( " << noutput_items << " * " << m_bytes_per_sample << " ) more than the buffer size ( " << m_nBufferSize << " ) !" << std::endl;
				noutput_items = int(m_nBufferSize/m_bytes_per_sample);
			}

retry_notify:

			while ((m_bBuffering) || (m_nBufferItems <= ((uint32_t)(m_fBufferLevel * (float)m_nBufferSize) + m_recv_samples_per_packet))) {
				// If getting too full, send them all through
				bool notified = true;
				if (m_bBuffering) {
					// Always wait for new samples to arrive while buffering
					m_hPacketEvent.wait(lock);			
				} else {
					xtime_get(&m_wait_next, CLOCK_MONOTONIC);
					m_wait_next.nsec += m_wait_delay.nsec;
					if (m_wait_next.nsec >= 1000000000) {
						m_wait_next.sec += 1;
						m_wait_next.nsec -= 1000000000;
					}
					// Wait for more samples to arrive, 
					// or wait just longer than it would have actually taken and use buffer samples
					notified = m_hPacketEvent.timed_wait(lock, m_wait_next);	
				}

				// Timeout
				if (notified == false) {
					std::cout << "rT" << std::endl;
					// Running late, use up some of the buffer
					break;								
				}

				if (m_bRunning == false) {
					std::cout << "No longer running after packet notification - signalling EOF" << std::endl;
					return -1;
				}

				// No longer filling buffer, so send samples back to runtime
				if (m_bBuffering == false) {
					break;
				}

				std::cout << "Caught packet signal while buffering!" << std::endl;
			}

			if (m_nBufferItems < m_recv_samples_per_packet) {
				std::cout << "rU" << std::endl;
				m_bBuffering = true;
				++m_nBufferUnderrunCount;

				// Keep waiting for buffer to fill back up sufficiently
				goto retry_notify;						
			} else if (m_nBufferItems < (noutput_items*m_bytes_per_sample))	{
				// Double check
				noutput_items = int(m_nBufferItems/m_bytes_per_sample);
			}


			++m_nReadPacketCount;
			uint32_t nPart1 = std::min(uint32_t(noutput_items*m_bytes_per_sample), m_nBufferSize - m_nBufferStart);
			uint8_t* p = m_pUSBBuffer + m_nBufferStart;
			uint32_t nResidual = (noutput_items*m_bytes_per_sample) - nPart1;
			//printf( "nPart1=%d\n", nPart1/m_bytes_per_sample);
			int q=nPart1/m_bytes_per_sample;
			if(m_bytes_per_sample==2) {
				float c=1/8.0/1024.0;
				for( int j=0; j<q; j++ ) {
					short r=(short)((uint16_t *)p)[j];
					out[j]=r*c;
				}
			} else {
				float c=1/256.0/1024.0/1024.0;
				for( int j=0; j<q; j++ ) {
					int r=(int)((uint32_t *)p)[j];
					out[j]=r*c;
				}
			}
//			memcpy(out, p, nPart1);

			if (nResidual) {
				int k=nPart1/m_bytes_per_sample;
				int q=nResidual/m_bytes_per_sample;
				if(m_bytes_per_sample==2) {
					float c=1/8.0/1024.0;
					for( int j=0; j<q; j++ ) {
						short r=(short)((uint16_t *)m_pUSBBuffer)[j];
						out[k+j]=r*c;
					}
				} else {
					float c=1/256.0/1024.0/1024.0;
					for( int j=0; j<q; j++ ) {
						int r=(int)((uint32_t *)m_pUSBBuffer)[j];
						out[k+j]=r*c;
					}
				}
//				memcpy(out + nPart1, m_pUSBBuffer, nResidual);
			}
		  
			m_nBufferItems -= noutput_items*m_bytes_per_sample;

			if (m_nBufferItems > 0) {
				m_nBufferStart = (m_nBufferStart + noutput_items*m_bytes_per_sample) % m_nBufferSize;
			}
		  

			// Tell runtime system how many output items we produced.
			return noutput_items/2;				
		}	

	} /* namespace elad */
} /* namespace gr */
